from typing import Type, Any, Optional, overload

from .core import resource as res
from .core.internal_models import meta_v1, autoscaling_v1

__all__ = ['create_global_resource', 'create_namespaced_resource']

_created_resources = {}


def get_generic_resource(version, kind):
    global _created_resources
    model = _created_resources.get((version, kind))
    return model[0] if model is not None else None


class Generic(dict):
    @overload
    def __init__(self, apiVersion: str=None, kind: str=None,
                 metadata: meta_v1.ObjectMeta=None, **kwargs):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def apiVersion(self) -> str:
        return self.get('apiVersion')

    @property
    def kind(self) -> str:
        return self.get('kind')

    @property
    def status(self) -> str:
        return self.get('status')

    @property
    def metadata(self) -> Optional[meta_v1.ObjectMeta]:
        meta = self.get('metadata')
        if meta is None:
            return None
        elif isinstance(meta, meta_v1.ObjectMeta):
            return meta
        return meta_v1.ObjectMeta.from_dict(meta)

    def __getattr__(self, item):
        if item.startswith("_"):
            raise AttributeError(f"{item} not found")
        return self.get(item)

    @classmethod
    def from_dict(cls, d: dict, lazy=True):
        return cls(d)

    def to_dict(self, dict_factory=dict):
        d = dict_factory(self)
        if 'metadata' in d and isinstance(d['metadata'], meta_v1.ObjectMeta):
            d['metadata'] = d['metadata'].to_dict(dict_factory)
        return d


def create_api_info(group, version, kind, plural, verbs=None) -> res.ApiInfo:
    if verbs is None:
        verbs = ['delete', 'deletecollection', 'get', 'global_list', 'global_watch', 'list', 'patch',
               'post', 'put', 'watch']
    return res.ApiInfo(
        resource=res.ResourceDef(group, version, kind),
        plural=plural,
        verbs=verbs
    )


class GenericGlobalScale(res.GlobalSubResource, autoscaling_v1.Scale):
    pass


class GenericGlobalStatus(res.GlobalSubResource, Generic):
    pass


class GenericNamespacedScale(res.NamespacedResourceG, autoscaling_v1.Scale):
    pass


class GenericNamespacedStatus(res.NamespacedResourceG, Generic):
    pass


def _create_subresource(main_class, parent_info: res.ApiInfo, action):
    class TmpName(main_class):
        _api_info = res.ApiInfo(
            resource=parent_info.resource if action == 'status' else res.ResourceDef('autoscaling', 'v1', 'Scale'),
            parent=parent_info.resource,
            plural=parent_info.plural,
            verbs=['get', 'patch', 'put'],
            action=action,
        )

    TmpName.__name__ = TmpName.__qualname__ = f"{parent_info.resource.kind}{action.capitalize()}"
    return TmpName


class GenericGlobalResource(res.GlobalResource, Generic):
    Scale: Type[GenericGlobalScale]
    Status: Type[GenericGlobalStatus]


class GenericNamespacedResource(res.NamespacedResourceG, Generic):
    Scale: Type[GenericNamespacedScale]
    Status: Type[GenericNamespacedStatus]


def _create_resource(namespaced, group, version, kind, plural, verbs=None) -> Any:
    global _created_resources
    res_key = (f'{group}/{version}', kind)
    signature = (namespaced, plural, tuple(verbs) if verbs else None)
    if res_key in _created_resources:
        model, curr_signature = _created_resources[res_key]
        if curr_signature != signature:
            raise ValueError(f"Resource {kind} already created but with different signature")
        return model

    if namespaced:
        main, status, scale = GenericNamespacedResource, GenericNamespacedStatus, GenericNamespacedScale
    else:
        main, status, scale = GenericGlobalResource, GenericGlobalStatus, GenericGlobalScale

    class TmpName(main):
        _api_info = create_api_info(group, version, kind, plural, verbs=verbs)

        Scale = _create_subresource(scale, _api_info, action='scale')
        Status = _create_subresource(status, _api_info, action='status')

    TmpName.__name__ = TmpName.__qualname__ = kind
    _created_resources[res_key] = (TmpName, signature)
    return TmpName


def create_global_resource(group: str, version: str, kind: str, plural: str, verbs=None) \
        -> Type[GenericGlobalResource]:
    """Create a new class representing a global resource with the provided specifications.

    **Parameters**

    * **group** `str` - API group of the resource. Example `stable.example.com`.
    * **version** `str` - API group version. Example `v1`.
    * **kind** `str` - Resource name. Example `Job`.
    * **plural** `str` - Resource collection name. Example `jobs`.

    **returns**  Subclass of `GenericGlobalResource`.
    """
    return _create_resource(
        False, group, version, kind, plural, verbs=verbs)


def create_namespaced_resource(group: str, version: str, kind: str, plural: str, verbs=None) \
        -> Type[GenericNamespacedResource]:
    """Create a new class representing a namespaced resource with the provided specifications.

    **Parameters**

    * **group** `str` - API group of the resource. Example `stable.example.com`.
    * **version** `str` - API group version. Example `v1`.
    * **kind** `str` - Resource name. Example `Job`.
    * **plural** `str` - Resource collection name. Example `jobs`.

    **returns**  Subclass of `GenericNamespacedResource`.
    """
    return _create_resource(
        True, group, version, kind, plural, verbs=verbs)



import time
from typing import Type, Any, Dict, Union
import dataclasses
from dataclasses import dataclass
import json
import asyncio

import httpx

from . import resource as r
from ..config.kubeconfig import KubeConfig, SingleConfig, DEFAULT_KUBECONFIG
from ..config import client_adapter
from .exceptions import ApiError
from .selector import build_selector
from ..types import OnErrorAction, OnErrorHandler, on_error_raise, PatchType


ALL_NS = '*'


def transform_exception(e: httpx.HTTPError):
    if isinstance(e, httpx.HTTPStatusError) and e.response.headers['Content-Type'] == 'application/json':
        return ApiError(request=e.request, response=e.response)
    return e


METHOD_MAPPING = {
    'delete': 'DELETE',
    'deletecollection': 'DELETE',
    'get': 'GET',
    'global_list': 'GET',
    'global_watch': 'GET',
    'list': 'GET',
    'patch': 'PATCH',
    'post': 'POST',
    'put': 'PUT',
    'watch': 'GET'
}

@dataclass
class BasicRequest:
    method: str
    url: str
    response_type: Any
    params: Dict[str, str] = dataclasses.field(default_factory=dict)
    data: Any = None
    headers: Dict[str, str] = None


class WatchDriver:
    def __init__(self, br: BasicRequest, build_request, lazy):
        self._version = br.params.get('resourceVersion')
        self._convert = br.response_type.from_dict
        self._br = br
        self._build_request = build_request
        self._lazy = lazy

    def get_request(self, timeout):
        br = self._br
        if self._version is not None:
            br.params['resourceVersion'] = self._version
        return self._build_request(br.method, br.url, params=br.params, timeout=timeout)

    def process_one_line(self, line):
        line = json.loads(line)
        tp = line['type']
        obj = line['object']
        self._version = obj['metadata']['resourceVersion']
        return tp, self._convert(obj, lazy=self._lazy)


class GenericClient:
    AdapterClient = staticmethod(client_adapter.Client)

    def __init__(self, config: Union[SingleConfig, KubeConfig] = None, namespace: str = None,
                 timeout: httpx.Timeout = None, lazy=True, trust_env: bool = True, field_manager: str = None):
        self._timeout = httpx.Timeout(10) if timeout is None else timeout
        self._watch_timeout = httpx.Timeout(self._timeout)
        self._watch_timeout.read = None
        self._lazy = lazy
        if config is None and trust_env:
            config = KubeConfig.from_env().get()
        elif config is None and not trust_env:
            config = KubeConfig.from_file(DEFAULT_KUBECONFIG).get()
        elif isinstance(config, KubeConfig):
            config = config.get()

        self.config = config
        self._client = self.AdapterClient(config, timeout, trust_env=trust_env)
        self._field_manager = field_manager
        self.namespace = namespace if namespace else config.namespace

    def prepare_request(self, method, res: Type[r.Resource] = None, obj=None, name=None, namespace=None,
                        watch: bool = False, params: dict = None, headers: dict = None) -> BasicRequest:
        if params is not None:
            params = {k: v for k, v in params.items() if v is not None}
        else:
            params = {}
        if headers is not None:
            headers = {k: v for k, v in headers.items() if v is not None}
        data = None
        if res is None:
            if obj is None:
                raise ValueError("At least a resource or an instance of a resource need to be provided")
            res = obj.__class__

        namespaced = issubclass(res, (r.NamespacedResource, r.NamespacedSubResource))

        if namespace == ALL_NS:
            if not issubclass(res, r.NamespacedResourceG):
                raise ValueError(f"Class {res} doesn't support global {method}")
            if method not in ('list', 'watch'):
                raise ValueError("Only methods 'list' and 'watch' can be called for all namespaces")
            real_method = "global_watch" if watch else "global_" + method
        else:
            real_method = "watch" if watch else method

        api_info = r.api_info(res)
        if real_method not in api_info.verbs:
            if watch:
                raise ValueError(f"Resource '{res.__name__}' is not watchable")
            else:
                raise ValueError(f"method '{method}' not supported by resource '{res.__name__}'")

        if watch:
            params['watch'] = "true"

        if api_info.parent is None:
            base = api_info.resource
        else:
            base = api_info.parent

        if base.group == '':
            path = ["api", base.version]
        else:
            path = ["apis", base.group, base.version]

        if namespaced and namespace != ALL_NS:
            if method in ('post', 'put') and obj.metadata.namespace is not None:
                if namespace is None:
                    namespace = obj.metadata.namespace
                elif namespace != obj.metadata.namespace:
                    raise ValueError(f"The namespace value '{namespace}' differ from the "
                                     f"namespace in the object metadata '{obj.metadata.namespace}'")
            if namespace is None:
                namespace = self.namespace
            path.extend(["namespaces", namespace])

        if method in ('post', 'put', 'patch'):
            if self._field_manager is not None and 'fieldManager' not in params:
                params['fieldManager'] = self._field_manager
            if method == 'patch' and headers['Content-Type'] == PatchType.APPLY.value and 'fieldManager' not in params:
                raise ValueError('Parameter "field_manager" is required for PatchType.APPLY')
            if obj is None:
                raise ValueError("obj is required for post, put or patch")

            if method == 'patch' and not isinstance(obj, r.Resource):
                data = obj
            else:
                data = obj.to_dict()
                # The following block, ensures that apiVersion and kind are always set.
                # this is needed as k8s fails if this data are not provided for objects derived by CRDs (Issue #27)
                if 'apiVersion' not in data:
                    data['apiVersion'] = api_info.resource.api_version
                if 'kind' not in data:
                    data['kind'] = api_info.resource.kind

        path.append(api_info.plural)
        if method in ('delete', 'get', 'patch', 'put') or api_info.action:
            if name is None and method == 'put':
                name = obj.metadata.name
            if name is None:
                raise ValueError("resource name not defined")
            path.append(name)

        if api_info.action:
            path.append(api_info.action)

        http_method = METHOD_MAPPING[method]
        if http_method == 'DELETE':
            res = None

        return BasicRequest(method=http_method, url="/".join(path), params=params, response_type=res, data=data, headers=headers)

    @staticmethod
    def raise_for_status(resp):
        try:
            resp.raise_for_status()
        except httpx.HTTPError as e:
            raise transform_exception(e)

    def build_adapter_request(self, br: BasicRequest):
        return self._client.build_request(br.method, br.url, params=br.params, json=br.data, headers=br.headers)

    def handle_response(self, method, resp, br):
        self.raise_for_status(resp)
        res = br.response_type
        if res is None:
            # TODO: delete/deletecollection actions normally return a Status object, we may want to return it as well
            return
        data = resp.json()
        if method == 'list':
            if 'metadata' in data and data['metadata'].get('continue'):
                cont = True
                br.params['continue'] = data['metadata']['continue']
            else:
                cont = False
            return cont, (res.from_dict(obj, lazy=self._lazy) for obj in data['items'])
        else:
            if res is not None:
                return res.from_dict(data, lazy=self._lazy)


class GenericSyncClient(GenericClient):
    def send(self, req, stream=False):
        return self._client.send(req, stream=stream)

    def watch(self, br: BasicRequest, on_error: OnErrorHandler = on_error_raise):
        wd = WatchDriver(br, self._client.build_request, self._lazy)
        err_count = 0
        while True:
            req = wd.get_request(timeout=self._watch_timeout)
            resp = self.send(req, stream=True)
            try:
                resp.raise_for_status()
                err_count = 0
                for line in resp.iter_lines():
                    yield wd.process_one_line(line)
            except Exception as e:
                err_count += 1
                handle_error = on_error(e, err_count)
                if handle_error.action is OnErrorAction.RAISE:
                    raise
                if handle_error.action is OnErrorAction.STOP:
                    break
                if handle_error.sleep > 0:
                    time.sleep(handle_error.sleep)
                continue

    def request(self, method, res: Type[r.Resource] = None, obj=None, name=None, namespace=None, watch: bool = False,
                headers: dict = None, params: dict = None) -> Any:
        br = self.prepare_request(method, res, obj, name, namespace, watch, headers=headers, params=params)
        req = self.build_adapter_request(br)
        resp = self.send(req)
        return self.handle_response(method, resp, br)

    def list(self, br: BasicRequest) -> Any:
        cont = True
        while cont:
            req = self.build_adapter_request(br)
            resp = self.send(req)
            cont, chunk = self.handle_response('list', resp, br)
            yield from chunk


class GenericAsyncClient(GenericClient):
    AdapterClient = staticmethod(client_adapter.AsyncClient)

    async def send(self, req, stream=False):
        return await self._client.send(req, stream=stream)

    async def watch(self, br: BasicRequest, on_error: OnErrorHandler = on_error_raise):
        wd = WatchDriver(br, self._client.build_request, self._lazy)
        err_count = 0
        while True:
            req = wd.get_request(timeout=self._watch_timeout)
            resp = await self.send(req, stream=True)
            try:
                resp.raise_for_status()
                err_count = 0
                async for line in resp.aiter_lines():
                    yield wd.process_one_line(line)
            except Exception as e:
                err_count += 1
                handle_error = on_error(e, err_count)
                if handle_error.action is OnErrorAction.RAISE:
                    raise
                if handle_error.action is OnErrorAction.STOP:
                    break
                if handle_error.sleep > 0:
                    await asyncio.sleep(handle_error.sleep)
                continue
            finally:
                await resp.aclose()

    async def request(self, method, res: Type[r.Resource] = None, obj=None, name=None, namespace=None,
                      watch: bool = False, headers: dict = None, params: dict = None) -> Any:
        br = self.prepare_request(method, res, obj, name, namespace, watch, headers=headers, params=params)
        req = self.build_adapter_request(br)
        resp = await self.send(req)
        return self.handle_response(method, resp, br)

    async def list(self, br: BasicRequest) -> Any:
        cont = True
        while cont:
            req = self.build_adapter_request(br)
            resp = await self.send(req)
            cont, chunk = self.handle_response('list', resp, br)
            for item in chunk:
                yield item

    async def close(self):
        await self._client.aclose()



import importlib
from typing import Union, TextIO, List

import yaml

from .generic_resource import get_generic_resource, GenericGlobalResource, GenericNamespacedResource
from .core.exceptions import LoadResourceError

try:
    import jinja2
except ImportError:
    jinja2 = None

REQUIRED_ATTR = ('apiVersion', 'kind')

AnyResource = Union[GenericGlobalResource, GenericNamespacedResource]


def _load_model(version, kind):
    if "/" in version:
        group, version_n = version.split("/")
        # Check if a generic resource was defined
        model = get_generic_resource(version, kind)
        if model is not None:
            return model

        # Generic resource not defined, but it could be a k8s resource
        if group.endswith(".k8s.io"):
            group = group[:-7]
        group = group.replace(".", "_")
        version = "_".join([group, version_n])
    else:
        version = f'core_{version}'

    try:
        module = importlib.import_module(f'lightkube.resources.{version.lower()}')
    except ImportError as e:
        # It was not a k8s resource and a generic resource was not previously defined
        raise LoadResourceError(f"{e}. If using a CRD, ensure you define a generic resource.")
    return getattr(module, kind)


def from_dict(d: dict) -> AnyResource:
    """Converts a kubernetes resource defined as python dict to the corresponding resource object.
    If the dict represent a standard resource, the function will automatically load the appropriate
    resource type. Generic resources are also supported and used assuming they were defined prior to
    the function call. Returns the resource object or raise a `LoadResourceError`.

    **parameters**

    * **d** - A dictionary representing a Kubernetes resource. Keys `apiVersion` and `kind` are
      always required.
    """
    for attr in REQUIRED_ATTR:
        if attr not in d:
            raise LoadResourceError(f"Invalid resource definition, key '{attr}' missing.")

    model = _load_model(d['apiVersion'], d['kind'])
    return model.from_dict(d)


def load_all_yaml(stream: Union[str, TextIO], context: dict = None, template_env = None) -> List[AnyResource]:
    """Load kubernetes resource objects defined as YAML. See `from_dict` regarding how resource types are detected.
    Returns a list of resource objects or raise a `LoadResourceError`.

    **parameters**

    * **stream** - A file-like object or a string representing a yaml file or a template resulting in
        a yaml file.
    * **context** - When is not `None` the stream is considered a `jinja2` template and the `context`
        will be used during templating.
    * **template_env** - `jinja2` template environment to be used for templating. When absent a standard
        environment is used.

    **NOTE**: When using the template functionality (setting the context parameter), the dependency
        module `jinja2` need to be installed.
    """
    if context is not None:
        stream = _template(stream, context=context, template_env=template_env)
    res = []
    for obj in yaml.safe_load_all(stream):
         res.append(from_dict(obj))
    return res


def dump_all_yaml(resources: List[AnyResource], stream: TextIO = None, indent=2):
    """Write kubernetes resource objects as YAML into an open file.

    **parameters**

    * **resources** - List of resources to write on the file
    * **stream** - Path to a file where to write the resources. When not set the content is returned
      as a string.
    * **indent** - Number of characters for indenting nasted blocks.
    """
    res = [r.to_dict() for r in resources]
    return yaml.safe_dump_all(res, stream, indent=indent)


def _template(stream: Union[str, TextIO], context: dict = None, template_env = None) -> List[AnyResource]:
    """
    Template a stream using jinja2 and the given context
    """
    if jinja2 is None:
        raise ImportError("load_from_template requires jinja2 to be installed")

    if template_env is None:
        template_env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
    elif not isinstance(template_env, jinja2.Environment):
        raise LoadResourceError("template_env is not a valid jinja2 template")

    tmpl = template_env.from_string(stream if isinstance(stream, str) else stream.read())
    return tmpl.render(**context)



"""
Exceptions.
"""
import httpx

from .internal_models import meta_v1


class ConfigError(Exception):
    """
    Configuration specific errors.
    """
    pass


class ApiError(httpx.HTTPStatusError):
    status: 'meta_v1.Status'

    def __init__(
            self, request: httpx.Request = None, response: httpx.Response = None) -> None:
        self.status = meta_v1.Status.from_dict(response.json())
        super().__init__(self.status.message, request=request, response=response)


class LoadResourceError(Exception):
    """
    Error in loading a resource
    """


class ObjectDeleted(Exception):
    """
    Object was unexpectedly deleted
    """

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"{self.name} was unexpectedly deleted"


class ConditionError(Exception):
    """
    Object is in specified bad condition
    """

    def __init__(self, name, messages):
        self.name = name
        self.messages = messages

    def __str__(self):
        messages = '; '.join(self.messages)
        return f'{self.name} has failure condition(s): {messages}'



from typing import Type, Iterator, TypeVar, Union, overload, Dict, Tuple, List, Iterable, AsyncIterable
import httpx
from ..config.kubeconfig import SingleConfig, KubeConfig
from .. import operators
from ..core import resource as r
from .generic_client import GenericSyncClient, GenericAsyncClient
from ..core.exceptions import ConditionError, ObjectDeleted
from ..types import OnErrorHandler, PatchType, on_error_raise
from .internal_resources import core_v1
from .selector import build_selector

NamespacedResource = TypeVar('NamespacedResource', bound=r.NamespacedResource)
GlobalResource = TypeVar('GlobalResource', bound=r.GlobalResource)
GlobalSubResource = TypeVar('GlobalSubResource', bound=r.GlobalSubResource)
NamespacedSubResource = TypeVar('NamespacedSubResource', bound=r.NamespacedSubResource)
AllNamespacedResource = TypeVar('AllNamespacedResource', r.NamespacedResource, r.NamespacedSubResource)
Resource = TypeVar('Resource', bound=r.Resource)
LabelValue = Union[str, None, operators.Operator, Iterable]
FieldValue = Union[str, operators.BinaryOperator, operators.SequenceOperator]
LabelSelector = Dict[str, LabelValue]
FieldSelector = Dict[str, FieldValue]


class Client:
    """Creates a new lightkube client

    **parameters**

    * **config** - Instance of `SingleConfig` or `KubeConfig`. When not set the configuration will be detected automatically
      using the following order: in-cluster config, `KUBECONFIG` environment variable, `~/.kube/config` file.
    * **namespace** - Default namespace to use. This attribute is used in case namespaced resources are called without
      defining a namespace. If not specified, the default namespace set in your kube configuration will be used.
    * **timeout** - Instance of `httpx.Timeout`. By default all timeouts are set to 10 seconds. Notice that read timeout
      is ignored when watching changes.
    * **lazy** - When set, the returned objects will be decoded from the JSON payload in a lazy way, i.e. only when
      accessed.
    * **field_manager** - Name associated with the actor or entity that is making these changes.
    * **trust_env** - Ignore environment variables, also passed through to httpx.Client trust_env.  See its
      docs for further description. If False, empty config will be derived from_file(DEFAULT_KUBECONFIG)
    """
    def __init__(self, config: Union[SingleConfig, KubeConfig] = None, namespace: str = None,
                 timeout: httpx.Timeout = None, lazy=True, field_manager: str = None, trust_env: bool = True):
        self._client = GenericSyncClient(config, namespace=namespace, timeout=timeout, lazy=lazy,
                                         field_manager=field_manager, trust_env=trust_env)

    @property
    def namespace(self):
        """Return the default namespace that will be used when a namespace has not been specified"""
        return self._client.namespace

    @property
    def config(self) -> SingleConfig:
        """Return the kubernetes configuration used in this client"""
        return self._client.config

    @overload
    def delete(self, res: Type[GlobalResource], name: str) -> None:
        ...

    @overload
    def delete(self, res: Type[NamespacedResource], name: str, *, namespace: str = None) -> None:
        ...

    def delete(self, res, name: str, *, namespace: str = None):
        """Delete an object

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to delete.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        """
        return self._client.request("delete", res=res, name=name, namespace=namespace)

    @overload
    def deletecollection(self, res: Type[GlobalResource]) -> None:
        ...

    @overload
    def deletecollection(self, res: Type[NamespacedResource], *, namespace: str = None) -> None:
        ...

    def deletecollection(self, res, *, namespace: str = None):
        """Delete all objects of the given kind

        * **res** - Resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        """
        return self._client.request("deletecollection", res=res, namespace=namespace)

    @overload
    def get(self, res: Type[GlobalResource], name: str) -> GlobalResource:
        ...

    @overload
    def get(self, res: Type[AllNamespacedResource], name: str, *, namespace: str = None) -> AllNamespacedResource:
        ...

    def get(self, res, name, *, namespace=None):
        """Return an object

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to fetch.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        """
        return self._client.request("get", res=res, name=name, namespace=namespace)

    @overload
    def list(self, res: Type[GlobalResource], *, chunk_size: int = None, labels: LabelSelector = None, fields: FieldSelector = None) -> \
            Iterator[GlobalResource]:
        ...

    @overload
    def list(self, res: Type[NamespacedResource], *, namespace: str = None, chunk_size: int = None,
             labels: LabelSelector = None, fields: FieldSelector = None) -> \
            Iterator[NamespacedResource]:
        ...

    def list(self, res, *, namespace=None, chunk_size=None, labels=None, fields=None):
        """Return an iterator of objects matching the selection criteria.

        **parameters**

        * **res** - resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **chunk_size** - *(optional)* Limit the amount of objects returned for each rest API call.
             This method will automatically execute all subsequent calls until no more data is available.
        * **labels** - *(optional)* Limit the returned objects by labels. More [details](../selectors).
        * **fields** - *(optional)* Limit the returned objects by fields. More [details](../selectors).
        """

        br = self._client.prepare_request(
            'list', res=res, namespace=namespace,
            params={
                'limit': chunk_size,
                'labelSelector': build_selector(labels) if labels else None,
                'fieldSelector': build_selector(fields, for_fields=True) if fields else None
            }
        )
        return self._client.list(br)

    @overload
    def watch(self, res: Type[GlobalResource], *, labels: LabelSelector = None, fields: FieldSelector = None,
              server_timeout: int = None,
              resource_version: str = None, on_error: OnErrorHandler = on_error_raise) -> \
            Iterator[Tuple[str, GlobalResource]]:
        ...

    @overload
    def watch(self, res: Type[NamespacedResource], *, namespace: str = None,
              labels: LabelSelector = None, fields: FieldSelector = None,
              server_timeout: int = None, resource_version: str = None,
              on_error: OnErrorHandler = on_error_raise) -> \
            Iterator[Tuple[str, NamespacedResource]]:
        ...

    def watch(self, res, *, namespace=None, labels=None, fields=None, server_timeout=None, resource_version=None, on_error=on_error_raise):
        """Watch changes to objects

        **parameters**

        * **res** - resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **labels** - *(optional)* Limit the returned objects by labels. More [details](../selectors).
        * **fields** - *(optional)* Limit the returned objects by fields. More [details](../selectors).
        * **server_timeout** - *(optional)* Server side timeout in seconds to close a watch request.
            This method will automatically create a new request whenever the backend close the connection
            without errors.
        * **resource_version** - *(optional)* When set, only modification events following this version will be returned.
        * **on_error** - *(optional)* Function that control what to do in case of errors.
            The default implementation will raise any error.
        """
        br = self._client.prepare_request("list", res=res, namespace=namespace, watch=True,
            params={
                'timeoutSeconds': server_timeout,
                'resourceVersion': resource_version,
                'labelSelector': build_selector(labels) if labels else None,
                'fieldSelector': build_selector(fields, for_fields=True) if fields else None
            }
        )
        return self._client.watch(br, on_error=on_error)

    @overload
    def wait(
        self,
        res: Type[GlobalResource],
        name: str,
        *,
        for_conditions: Iterable[str],
        raise_for_conditions: Iterable[str] = (),
    ) -> GlobalResource:
        ...

    @overload
    def wait(
        self,
        res: Type[AllNamespacedResource],
        name: str,
        *,
        for_conditions: Iterable[str],
        namespace: str = None,
        raise_for_conditions: Iterable[str] = (),
    ) -> AllNamespacedResource:
        ...

    def wait(
        self,
        res,
        name: str,
        *,
        for_conditions: Iterable[str],
        namespace=None,
        raise_for_conditions: Iterable[str] = (),
    ):
        """Waits for specified conditions.

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of resource to wait for.
        * **for_conditions** - Condition types that are considered a success and will end the wait.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **raise_for_conditions** - *(optional)* Condition types that are considered failures and will exit the wait early.
        """

        kind = r.api_info(res).plural
        full_name = f'{kind}/{name}'

        for_conditions = list(for_conditions)
        raise_for_conditions = list(raise_for_conditions)

        for op, obj in self.watch(res, namespace=namespace, fields={'metadata.name': name}):
            if obj.status is None:
                continue

            if op == "DELETED":
                raise ObjectDeleted(full_name)

            try:
                status = obj.status.to_dict()
            except AttributeError:
                status = obj.status

            conditions = [c for c in status.get('conditions', []) if c['status'] == 'True']
            if any(c['type'] in for_conditions for c in conditions):
                return obj

            failures = [c for c in conditions if c['type'] in raise_for_conditions]

            if failures:
                raise ConditionError(full_name, [f.get('message', f['type']) for f in failures])

    @overload
    def patch(self, res: Type[GlobalSubResource], name: str,
              obj: Union[GlobalSubResource, Dict, List], *,
              patch_type: PatchType = PatchType.STRATEGIC,
              field_manager: str = None, force: bool = False) -> GlobalSubResource:
        ...

    @overload
    def patch(self, res: Type[GlobalResource], name: str,
              obj: Union[GlobalResource, Dict, List], *,
              patch_type: PatchType = PatchType.STRATEGIC,
              field_manager: str = None, force: bool = False) -> GlobalResource:
        ...

    @overload
    def patch(self, res: Type[AllNamespacedResource], name: str,
              obj: Union[AllNamespacedResource, Dict, List], *, namespace: str = None,
              patch_type: PatchType = PatchType.STRATEGIC,
              field_manager: str = None, force: bool = False) -> AllNamespacedResource:
        ...

    def patch(self, res, name, obj, *, namespace=None, patch_type=PatchType.STRATEGIC, field_manager=None, force=False):
        """Patch an object.

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to patch.
        * **obj** - patch object.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **patch_type** - *(optional)* Type of patch to execute. Default `PatchType.STRATEGIC`.
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
            **NOTE**: This parameter is mandatory (here or at `Client` creation time) for `PatchType.APPLY`.
        * **force** - *(optional)* Force is going to "force" Apply requests. It means user will re-acquire conflicting
          fields owned by other people. This parameter is ignored for non-apply patch types
        """
        force_param = 'true' if force and patch_type == PatchType.APPLY else None
        return self._client.request("patch", res=res, name=name, namespace=namespace, obj=obj,
                                    headers={'Content-Type': patch_type.value},
                                    params={'force': force_param, 'fieldManager': field_manager})



    @overload
    def create(self, obj: GlobalSubResource,  name: str, field_manager: str = None) -> GlobalSubResource:
        ...

    @overload
    def create(self, obj: NamespacedSubResource, name: str, *, namespace: str = None, field_manager: str = None) \
            -> NamespacedSubResource:
        ...

    @overload
    def create(self, obj: GlobalResource, field_manager: str = None) -> GlobalResource:
        ...

    @overload
    def create(self, obj: NamespacedResource, field_manager: str = None) -> NamespacedResource:
        ...

    def create(self, obj, name=None, *, namespace=None, field_manager=None):
        """Creates a new object

        **parameters**

        * **obj** - object to create. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
        """
        return self._client.request("post", name=name, namespace=namespace, obj=obj,
                                    params={'fieldManager': field_manager})

    @overload
    def replace(self, obj: GlobalSubResource, name: str, field_manager: str = None) -> GlobalSubResource:
        ...

    @overload
    def replace(self, obj: NamespacedSubResource, name: str, *, namespace: str = None, field_manager: str = None) \
            -> NamespacedSubResource:
        ...

    @overload
    def replace(self, obj: GlobalResource, field_manager: str = None) -> GlobalResource:
        ...

    @overload
    def replace(self, obj: NamespacedResource, field_manager: str = None) -> NamespacedResource:
        ...

    def replace(self, obj, name=None, *, namespace=None, field_manager=None):
        """Replace an existing resource.

        **parameters**

        * **obj** - new object. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
        """
        return self._client.request("put", name=name, namespace=namespace, obj=obj,
                                    params={'fieldManager': field_manager})

    @overload
    def log(self, name:str, *, namespace: str = None, container: str = None, follow: bool = False,
            since: int = None, tail_lines: int = None, timestamps: bool = False) -> Iterator[str]:
        ...

    def log(self, name, *, namespace=None, container=None, follow=False,
            since=None, tail_lines=None, timestamps=False):
        """Return log lines for the given Pod

        **parameters**

        * **name** - Name of the Pod.
        * **namespace** - *(optional)* Name of the namespace containing the Pod.
        * **container** - *(optional)* The container for which to stream logs. Defaults to only container if there is one container in the pod.
        * **follow** - *(optional)* If `True`, follow the log stream of the pod.
        * **since** - *(optional)* If set, a relative time in seconds before the current time from which to fetch logs.
        * **tail_lines** - *(optional)* If set, the number of lines from the end of the logs to fetch.
        * **timestamps** - *(optional)* If `True`, add an RFC3339 or RFC3339Nano timestamp at the beginning of every line of log output.
        """
        br = self._client.prepare_request(
            'get', core_v1.PodLog, name=name, namespace=namespace,
            params={'timestamps': timestamps, 'tailLines': tail_lines, 'container': container,
                    'sinceSeconds': since, 'follow': follow})
        req = self._client.build_adapter_request(br)
        resp = self._client.send(req, stream=follow)
        self._client.raise_for_status(resp)
        return resp.iter_lines()

    @overload
    def apply(self, obj: GlobalSubResource,  name: str, *, field_manager: str = None, force: bool = False) \
            -> GlobalSubResource:
        ...

    @overload
    def apply(self, obj: NamespacedSubResource, name: str, *, namespace: str = None,
              field_manager: str = None, force: bool = False) -> NamespacedSubResource:
        ...

    @overload
    def apply(self, obj: GlobalResource, field_manager: str = None, force: bool = False) -> GlobalResource:
        ...

    @overload
    def apply(self, obj: NamespacedResource, field_manager: str = None, force: bool = False) -> NamespacedResource:
        ...

    def apply(self, obj, name=None, *, namespace=None, field_manager=None, force=False):
        """Create or configure an object. This method uses the
        [server-side apply](https://kubernetes.io/docs/reference/using-api/server-side-apply/) functionality.

        **parameters**

        * **obj** - object to create. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - Name associated with the actor or entity that is making these changes.
        * **force** - *(optional)* Force is going to "force" Apply requests. It means user will re-acquire conflicting
          fields owned by other people.
        """
        if namespace is None and isinstance(obj, r.NamespacedResource) and obj.metadata.namespace:
            namespace = obj.metadata.namespace
        if name is None and obj.metadata.name:
            name = obj.metadata.name
        return self.patch(type(obj), name, obj, namespace=namespace,
                          patch_type=PatchType.APPLY, field_manager=field_manager, force=force)


class AsyncClient:
    """Creates a new lightkube client

    **parameters**

    * **config** - Instance of `SingleConfig` or `KubeConfig`. When not set the configuration will be detected automatically
      using the following order: in-cluster config, `KUBECONFIG` environment variable, `~/.kube/config` file.
    * **namespace** - Default namespace to use. This attribute is used in case namespaced resources are called without
      defining a namespace. If not specified, the default namespace set in your kube configuration will be used.
    * **timeout** - Instance of `httpx.Timeout`. By default all timeouts are set to 10 seconds. Notice that read timeout
      is ignored when watching changes.
    * **lazy** - When set, the returned objects will be decoded from the JSON payload in a lazy way, i.e. only when
      accessed.
    * **field_manager** - Name associated with the actor or entity that is making these changes.
    * **trust_env** - Ignore environment variables, also passed through to httpx.AsyncClient trust_env.  See its
      docs for further description. If False, empty config will be derived from_file(DEFAULT_KUBECONFIG)
    """
    def __init__(self, config: Union[SingleConfig, KubeConfig] = None, namespace: str = None,
                 timeout: httpx.Timeout = None, lazy=True, field_manager: str = None, trust_env: bool = True):
        self._client = GenericAsyncClient(config, namespace=namespace, timeout=timeout, lazy=lazy,
                                          field_manager=field_manager, trust_env=trust_env)

    @property
    def namespace(self):
        """Return the default namespace that will be used when a namespace has not been specified"""
        return self._client.namespace

    @property
    def config(self) -> SingleConfig:
        """Return the kubernetes configuration used in this client"""
        return self._client.config

    @overload
    async def delete(self, res: Type[GlobalResource], name: str) -> None:
        ...

    @overload
    async def delete(self, res: Type[NamespacedResource], name: str, *, namespace: str = None) -> None:
        ...

    async def delete(self, res, name: str, *, namespace: str = None):
        """Delete an object

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to delete.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        """
        return await self._client.request("delete", res=res, name=name, namespace=namespace)

    @overload
    async def deletecollection(self, res: Type[GlobalResource]) -> None:
        ...

    @overload
    async def deletecollection(self, res: Type[NamespacedResource], *, namespace: str = None) -> None:
        ...

    async def deletecollection(self, res, *, namespace: str = None):
        """Delete all objects of the given kind

        * **res** - Resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        """
        return await self._client.request("deletecollection", res=res, namespace=namespace)

    @overload
    async def get(self, res: Type[GlobalResource], name: str) -> GlobalResource:
        ...

    @overload
    async def get(self, res: Type[AllNamespacedResource], name: str, *, namespace: str = None) -> AllNamespacedResource:
        ...

    async def get(self, res, name, *, namespace=None):
        """Return an object

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to fetch.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        """
        return await self._client.request("get", res=res, name=name, namespace=namespace)

    @overload
    def list(self, res: Type[GlobalResource], *, chunk_size: int = None, labels: LabelSelector = None, fields: FieldSelector = None) -> \
            AsyncIterable[GlobalResource]:
        ...

    @overload
    def list(self, res: Type[NamespacedResource], *, namespace: str = None, chunk_size: int = None,
             labels: LabelSelector = None, fields: FieldSelector = None) -> \
            AsyncIterable[NamespacedResource]:
        ...

    def list(self, res, *, namespace=None, chunk_size=None, labels=None, fields=None):
        """Return an iterator of objects matching the selection criteria.

        **parameters**

        * **res** - resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **chunk_size** - *(optional)* Limit the amount of objects returned for each rest API call.
             This method will automatically execute all subsequent calls until no more data is available.
        * **labels** - *(optional)* Limit the returned objects by labels. More [details](../selectors).
        * **fields** - *(optional)* Limit the returned objects by fields. More [details](../selectors).
        """

        br = self._client.prepare_request(
            'list', res=res, namespace=namespace,
            params={
                'limit': chunk_size,
                'labelSelector': build_selector(labels) if labels else None,
                'fieldSelector': build_selector(fields, for_fields=True) if fields else None
            }
        )
        return self._client.list(br)

    @overload
    def watch(self, res: Type[GlobalResource], *, labels: LabelSelector = None, fields: FieldSelector = None,
              server_timeout: int = None,
              resource_version: str = None, on_error: OnErrorHandler = on_error_raise) -> \
            AsyncIterable[Tuple[str, GlobalResource]]:
        ...

    @overload
    def watch(self, res: Type[NamespacedResource], *, namespace: str = None,
              labels: LabelSelector = None, fields: FieldSelector = None,
              server_timeout: int = None, resource_version: str = None,
              on_error: OnErrorHandler = on_error_raise) -> \
            AsyncIterable[Tuple[str, NamespacedResource]]:
        ...

    def watch(self, res, *, namespace=None, labels=None, fields=None, server_timeout=None, resource_version=None, on_error=on_error_raise):
        """Watch changes to objects

        **parameters**

        * **res** - resource kind.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **labels** - *(optional)* Limit the returned objects by labels. More [details](../selectors).
        * **fields** - *(optional)* Limit the returned objects by fields. More [details](../selectors).
        * **server_timeout** - *(optional)* Server side timeout in seconds to close a watch request.
            This method will automatically create a new request whenever the backend close the connection
            without errors.
        * **resource_version** - *(optional)* When set, only modification events following this version will be returned.
        * **on_error** - *(optional)* Function that control what to do in case of errors.
            The default implementation will raise any error.
        """
        br = self._client.prepare_request("list", res=res, namespace=namespace, watch=True,
            params={
                'timeoutSeconds': server_timeout,
                'resourceVersion': resource_version,
                'labelSelector': build_selector(labels) if labels else None,
                'fieldSelector': build_selector(fields, for_fields=True) if fields else None
            }
        )
        return self._client.watch(br, on_error=on_error)

    @overload
    async def wait(
        self,
        res: Type[GlobalResource],
        name: str,
        *,
        for_conditions: Iterable[str],
        raise_for_conditions: Iterable[str] = (),
    ) -> GlobalResource:
        ...

    @overload
    async def wait(
        self,
        res: Type[AllNamespacedResource],
        name: str,
        *,
        for_conditions: Iterable[str],
        namespace: str = None,
        raise_for_conditions: Iterable[str] = (),
    ) -> AllNamespacedResource:
        ...

    async def wait(
        self,
        res,
        name: str,
        *,
        for_conditions: Iterable[str],
        namespace=None,
        raise_for_conditions: Iterable[str] = (),
    ):
        """Waits for specified conditions.

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of resource to wait for.
        * **for_conditions** - Condition types that are considered a success and will end the wait.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **raise_for_conditions** - *(optional)* Condition types that are considered failures and will exit the wait early.
        """

        kind = r.api_info(res).plural
        full_name = f'{kind}/{name}'

        for_conditions = list(for_conditions)
        raise_for_conditions = list(raise_for_conditions)

        watch = self.watch(res, namespace=namespace, fields={'metadata.name': name})
        try:
            async for op, obj in watch:

                if obj.status is None:
                    continue

                if op == "DELETED":
                    raise ObjectDeleted(full_name)

                try:
                    status = obj.status.to_dict()
                except AttributeError:
                    status = obj.status

                conditions = [c for c in status.get('conditions', []) if c['status'] == 'True']
                if any(c['type'] in for_conditions for c in conditions):
                    return obj

                failures = [c for c in conditions if c['type'] in raise_for_conditions]

                if failures:
                    raise ConditionError(full_name, [f.get('message', f['type']) for f in failures])
        finally:
            # we ensure the async generator is closed before returning
            await watch.aclose()

    @overload
    async def patch(self, res: Type[GlobalSubResource], name: str,
                    obj: Union[GlobalSubResource, Dict, List], *,
                    patch_type: PatchType = PatchType.STRATEGIC,
                    field_manager: str = None, force: bool = False) -> GlobalSubResource:
        ...

    @overload
    async def patch(self, res: Type[GlobalResource], name: str,
                    obj: Union[GlobalResource, Dict, List], *,
                    patch_type: PatchType = PatchType.STRATEGIC,
                    field_manager: str = None, force: bool = False) -> GlobalResource:
        ...

    @overload
    async def patch(self, res: Type[AllNamespacedResource], name: str,
                    obj: Union[AllNamespacedResource, Dict, List], *, namespace: str = None,
                    patch_type: PatchType = PatchType.STRATEGIC,
                    field_manager: str = None, force: bool = False) -> AllNamespacedResource:
        ...

    async def patch(self, res, name, obj, *, namespace=None, patch_type=PatchType.STRATEGIC,
                    field_manager=None, force=False):
        """Patch an object.

        **parameters**

        * **res** - Resource kind.
        * **name** - Name of the object to patch.
        * **obj** - patch object.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **patch_type** - *(optional)* Type of patch to execute. Default `PatchType.STRATEGIC`.
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
            **NOTE**: This parameter is mandatory (here or at `Client` creation time) for `PatchType.APPLY`.
        * **force** - *(optional)* Force is going to "force" Apply requests. It means user will re-acquire conflicting
          fields owned by other people. This parameter is ignored for non-apply patch types
        """
        force_param = 'true' if force and patch_type == PatchType.APPLY else None
        return await self._client.request("patch", res=res, name=name, namespace=namespace, obj=obj,
                                          headers={'Content-Type': patch_type.value},
                                          params={'force': force_param, 'fieldManager': field_manager})

    @overload
    async def create(self, obj: GlobalSubResource,  name: str, field_manager: str = None) -> GlobalSubResource:
        ...

    @overload
    async def create(self, obj: NamespacedSubResource, name: str, *, namespace: str = None, field_manager: str = None) \
            -> NamespacedSubResource:
        ...

    @overload
    async def create(self, obj: GlobalResource, field_manager: str = None) -> GlobalResource:
        ...

    @overload
    async def create(self, obj: NamespacedResource, field_manager: str = None) -> NamespacedResource:
        ...

    async def create(self, obj, name=None, *, namespace=None, field_manager=None):
        """Creates a new object

        **parameters**

        * **obj** - object to create. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
        """
        return await self._client.request("post", name=name, namespace=namespace, obj=obj,
                                          params={'fieldManager': field_manager})

    @overload
    async def replace(self, obj: GlobalSubResource, name: str, field_manager: str = None) -> GlobalSubResource:
        ...

    @overload
    async def replace(self, obj: NamespacedSubResource, name: str, *, namespace: str = None,
                      field_manager: str = None) -> NamespacedSubResource:
        ...

    @overload
    async def replace(self, obj: GlobalResource, field_manager: str = None) -> GlobalResource:
        ...

    @overload
    async def replace(self, obj: NamespacedResource, field_manager: str = None) -> NamespacedResource:
        ...

    async def replace(self, obj, name=None, *, namespace=None, field_manager=None):
        """Replace an existing resource.

        **parameters**

        * **obj** - new object. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - *(optional)* Name associated with the actor or entity that is making these changes.
            This parameter overrides the corresponding `Client` initialization parameter.
        """
        return await self._client.request("put", name=name, namespace=namespace, obj=obj,
                                          params={'fieldManager': field_manager})

    @overload
    def log(self, name:str, *, namespace: str = None, container: str = None, follow: bool = False,
            since: int = None, tail_lines: int = None, timestamps: bool = False) -> AsyncIterable[str]:
        ...

    def log(self, name, *, namespace=None, container=None, follow=False,
            since=None, tail_lines=None, timestamps=False):
        """Return log lines for the given Pod

        **parameters**

        * **name** - Name of the Pod.
        * **namespace** - *(optional)* Name of the namespace containing the Pod.
        * **container** - *(optional)* The container for which to stream logs. Defaults to only container if there is one container in the pod.
        * **follow** - *(optional)* If `True`, follow the log stream of the pod.
        * **since** - *(optional)* If set, a relative time in seconds before the current time from which to fetch logs.
        * **tail_lines** - *(optional)* If set, the number of lines from the end of the logs to fetch.
        * **timestamps** - *(optional)* If `True`, add an RFC3339 or RFC3339Nano timestamp at the beginning of every line of log output.
        """
        br = self._client.prepare_request(
            'get', core_v1.PodLog, name=name, namespace=namespace,
            params={'timestamps': timestamps, 'tailLines': tail_lines, 'container': container,
                    'sinceSeconds': since, 'follow': follow})
        req = self._client.build_adapter_request(br)

        async def stream_log():
            resp = await self._client.send(req, stream=follow)
            self._client.raise_for_status(resp)
            async for line in resp.aiter_lines():
                yield line
        return stream_log()

    @overload
    async def apply(self, obj: GlobalSubResource,  name: str, *, field_manager: str = None, force: bool = False) \
            -> GlobalSubResource:
        ...

    @overload
    async def apply(self, obj: NamespacedSubResource, name: str, *, namespace: str = None,
              field_manager: str = None, force: bool = False) -> NamespacedSubResource:
        ...

    @overload
    async def apply(self, obj: GlobalResource, field_manager: str = None, force: bool = False) -> GlobalResource:
        ...

    @overload
    async def apply(self, obj: NamespacedResource, field_manager: str = None, force: bool = False) -> NamespacedResource:
        ...

    async def apply(self, obj, name=None, *, namespace=None, field_manager=None, force=False):
        """Create or configure an object. This method uses the
        [server-side apply](https://kubernetes.io/docs/reference/using-api/server-side-apply/) functionality.

        **parameters**

        * **obj** - object to create. This need to be an instance of a resource kind.
        * **name** - *(optional)* Required only for sub-resources: Name of the resource to which this object belongs.
        * **namespace** - *(optional)* Name of the namespace containing the object (Only for namespaced resources).
        * **field_manager** - Name associated with the actor or entity that is making these changes.
        * **force** - *(optional)* Force is going to "force" Apply requests. It means user will re-acquire conflicting
          fields owned by other people.
        """
        if namespace is None and isinstance(obj, r.NamespacedResource) and obj.metadata.namespace:
            namespace = obj.metadata.namespace
        if name is None and obj.metadata.name:
            name = obj.metadata.name
        return await self.patch(type(obj), name, obj, namespace=namespace,
                                patch_type=PatchType.APPLY, field_manager=field_manager, force=force)

    async def close(self):
        """Close the underline httpx client"""
        await self._client.close()



from typing import NamedTuple, List, Optional
from dataclasses import dataclass


class ResourceDef(NamedTuple):
    group: str
    version: str
    kind: str

    @property
    def api_version(self):
        return f"{self.group}/{self.version}" if self.group else self.version


@dataclass
class ApiInfo:
    resource: ResourceDef
    plural: str
    verbs: List[str]
    parent: Optional[ResourceDef] = None
    action: str = None


class Resource:
    _api_info: ApiInfo


def api_info(res: Resource):
    return res._api_info


class NamespacedResource(Resource):
    pass


class NamespacedSubResource(Resource):
    pass


class GlobalResource(Resource):
    pass


class NamespacedResourceG(NamespacedResource):
    pass


class GlobalSubResource(Resource):
    pass



import enum
from dataclasses import dataclass
import typing


class PatchType(enum.Enum):
    JSON = 'application/json-patch+json'
    MERGE = 'application/merge-patch+json'
    STRATEGIC = 'application/strategic-merge-patch+json'
    APPLY = 'application/apply-patch+yaml'


class OnErrorAction(enum.Enum):
    RETRY = 0       # retry to perform the API call again from the last version
    STOP = 1        # stop silently the iterator
    RAISE = 2       # raise the error on the caller scope


@dataclass
class OnErrorResult:
    action: OnErrorAction
    sleep: float = 0


OnErrorHandler = typing.Callable[[Exception, int], OnErrorResult]


def on_error_raise(e, count):
    return OnErrorResult(OnErrorAction.RAISE)


def on_error_stop(e, count):
    return OnErrorResult(OnErrorAction.STOP)


def on_error_retry(e, count):
    return OnErrorResult(OnErrorAction.RETRY)



import sys

try:
    from ..resources import core_v1
except:
    if sys.modules["__main__"].__package__ != 'mkdocs':   # we ignore this import error during documentation generation
        raise
    from unittest import mock

    class PodLog:
        pass

    core_v1 = mock.Mock()
    core_v1.PodLog = PodLog



from .core.client import Client, AsyncClient
from .core.generic_client import ALL_NS
from .core.exceptions import ApiError, ConfigError, LoadResourceError
from .config.kubeconfig import KubeConfig, SingleConfig



import sys

try:
    from ..models import meta_v1, autoscaling_v1
except:
    if sys.modules["__main__"].__package__ != 'mkdocs':   # we ignore this import error during documentation generation
        raise
    from unittest import mock

    class ObjectMeta:
        pass

    meta_v1 = mock.Mock()
    meta_v1.ObjectMeta = ObjectMeta


    class Scale:
        pass

    autoscaling_v1 = mock.Mock()
    autoscaling_v1.Scale = Scale



import os
import yaml
import tempfile
import base64
from typing import Dict, NamedTuple, Optional
from pathlib import Path

from ..core import exceptions
from .models import Cluster, User, Context

"""
| behavior                  | kubectl                   | lightkube             |
|---------------------------|---------------------------|-----------------------|
| current-context missing   | use proxy                 | fail (conf is None)   |
| current-context wrong     | fail                      | fail                  |
| context.cluster missing   | use proxy                 | fail                  |
| context.user missing      | interactive user/password | no auth set           |
| context.user wrong        | interactive user/password | fail                  |
| context.namespace missing | use default namespace     | use default namespace |
"""


def to_mapping(obj_list, key, factory):
    return {obj['name']: factory.from_dict(obj[key], lazy=False) for obj in obj_list}


DEFAULT_NAMESPACE = "default"
SERVICE_ACCOUNT = "/var/run/secrets/kubernetes.io/serviceaccount"
DEFAULT_KUBECONFIG = "~/.kube/config"


class SingleConfig(NamedTuple):
    context_name: str
    context: Context
    cluster: Cluster
    user: User = None
    fname: Path = None

    @property
    def namespace(self):
        return self.context.namespace or DEFAULT_NAMESPACE

    def abs_file(self, fname):
        if Path(fname).is_absolute():
            return fname

        if self.fname is None:
            raise exceptions.ConfigError(f"{fname} is relative, but kubeconfig path unknown")

        return self.fname.parent.joinpath(fname)


PROXY_CONF = SingleConfig(
    context_name="default", context=Context(cluster="default"),
    cluster=Cluster(server="http://localhost:8080")
)


class KubeConfig:
    """Class to represent a kubeconfig. See the specific constructors depending on your use case."""
    clusters: Dict[str, Cluster]
    users: Dict[str, User]
    contexts: Dict[str, Context]

    def __init__(self, *, clusters, contexts, users=None, current_context=None, fname=None):
        """
        Create the kubernetes configuration manually. Normally this constructor should not be called directly.
        Use a specific constructor instead.

        **Parameters**

        * **clusters**: Dictionary of cluster name -> `Cluster` instance.
        * **contexts**: Dictionary of context name -> `Context` instance.
        * **users**: Dictionary of user name -> `User` instance.
        * **current_context**: Name of the current context.
        * **fname**: Name of the file where the configuration has been readed from.
        """
        self.current_context = current_context
        self.clusters = clusters
        self.contexts = contexts
        self.users = users or {}
        self.fname = Path(fname) if fname else None

    @classmethod
    def from_dict(cls, conf: Dict, fname=None):
        """Creates a KubeConfig instance from the content of a dictionary structure.

        **Parameters**

        * **conf**: Configuration structure, main attributes are `clusters`, `contexts`, `users` and `current-context`.
        * **fname**: File path from where this configuration has been loaded. This is needed to resolve relative paths
          present inside the configuration.
        """
        return cls(
            current_context=conf.get('current-context'),
            clusters=to_mapping(conf['clusters'], 'cluster', factory=Cluster),
            contexts=to_mapping(conf['contexts'], 'context', factory=Context),
            users=to_mapping(conf.get('users', []), 'user', factory=User),
            fname=fname
        )

    def get(self, context_name=None, default: SingleConfig = None) -> Optional[SingleConfig]:
        """Returns a `SingleConfig` instance, representing the configuration matching the given `context_name`.
        Lightkube client will automatically call this method without parameters when an instance of `KubeConfig`
        is provided.

        **Parameters**

        * **context_name**: Name of the context to use. If `context_name` is undefined, the `current-context` is used.
          In the case both contexts are undefined, and the default is provided, this method will return the default.
          It will fail with an error otherwise.
        * **default**: Instance of a `SingleConfig` to be returned in case both contexts are not set. When this
          parameter is not provided and no context is defined, the method call will fail.
        """
        if context_name is None:
            context_name = self.current_context
        if context_name is None:
            if default is None:
                raise exceptions.ConfigError("No current context set and no default provided")
            return default
        try:
            ctx = self.contexts[context_name]
        except KeyError:
            raise exceptions.ConfigError(f"Context '{context_name}' not found")
        return SingleConfig(
            context_name=context_name, context=ctx,
            cluster=self.clusters[ctx.cluster],
            user=self.users[ctx.user] if ctx.user else None,
            fname=self.fname
        )

    @classmethod
    def from_file(cls, fname):
        """Creates an instance of the KubeConfig class from a kubeconfig file.

        **Parameters**

         * **fname**: Path to the kuberneted configuration.
        """
        filepath = Path(fname).expanduser()
        if not filepath.is_file():
            raise exceptions.ConfigError(f"Configuration file {fname} not found")
        with filepath.open() as f:
            return cls.from_dict(yaml.safe_load(f.read()), fname=filepath)

    @classmethod
    def from_one(cls, *, cluster, user=None, context_name='default', namespace=None, fname=None):
        """Creates an instance of the KubeConfig class from one cluster and one user configuration"""
        context = Context(cluster=context_name, user=context_name if user else None, namespace=namespace)
        return cls(
            clusters={context_name: cluster},
            contexts={context_name: context},
            users={context_name: user} if user else None,
            current_context=context_name,
            fname=fname
        )

    @classmethod
    def from_server(cls, url, namespace=None):
        """Creates an instance of the KubeConfig class from the cluster server url"""
        return cls.from_one(cluster=Cluster(server=url), namespace=namespace)

    @classmethod
    def from_service_account(cls, service_account=SERVICE_ACCOUNT):
        """Creates a configuration from in-cluster service account information.

        **Parameters**

         * **service_account**: Allows to override the default service account directory path.
           Default `/var/run/secrets/kubernetes.io/serviceaccount`.
        """
        account_dir = Path(service_account)

        try:
            token = account_dir.joinpath("token").read_text()
            namespace = account_dir.joinpath("namespace").read_text()
        except FileNotFoundError as e:
            raise exceptions.ConfigError(str(e))

        host = os.environ["KUBERNETES_SERVICE_HOST"]
        port = os.environ["KUBERNETES_SERVICE_PORT"]
        if ":" in host:     # ipv6
            host = f"[{host}]"
        return cls.from_one(
            cluster=Cluster(
                server=f"https://{host}:{port}",
                certificate_auth=str(account_dir.joinpath("ca.crt"))
            ),
            user=User(token=token),
            namespace=namespace
        )

    @classmethod
    def from_env(cls, service_account=SERVICE_ACCOUNT, default_config=DEFAULT_KUBECONFIG):
        """Attempts to load the configuration automatically looking at the environment and filesystem.

         The method will attempt to load a configuration using the following order:

         * in-cluster config.
         * config file defined in `KUBECONFIG` environment variable.
         * configuration file present on the default location.

         **Parameters**

         * **service_account**: Allows to override the default service account directory path.
           Default `/var/run/secrets/kubernetes.io/serviceaccount`.
         * **default_config**: Allows to override the default configuration location. Default `~/.kube/config`.
         """
        try:
            return KubeConfig.from_service_account(service_account=service_account)
        except exceptions.ConfigError:
            return KubeConfig.from_file(os.environ.get('KUBECONFIG', default_config))




from typing import Dict, Union, List
from collections.abc import Iterable
from lightkube import operators

FIELDS_SUPPORT = ('equal', 'not_equal', 'not_in')
FIELDS_SUPPORT_STR = ', '.join(f'"{fs}"' for fs in FIELDS_SUPPORT)


def build_selector(pairs: Union[List, Dict], for_fields=False):
    res = []
    if not isinstance(pairs, list):
        pairs = pairs.items()
    for k, v in pairs:
        if v is None:
            v = operators.exists()
        elif isinstance(v, str):
            v = operators.equal(v)
        elif isinstance(v, Iterable):
            v = operators.in_(v)

        if not isinstance(v, operators.Operator):
            raise ValueError(f"selector value '{v}' should be str, None, Iterable or instance of operator")

        if for_fields and v.op_name not in FIELDS_SUPPORT:
            raise ValueError(f"parameter 'fields' only support operators {FIELDS_SUPPORT_STR}")

        if for_fields and v.op_name == 'not_in':    # not_in can be implement using several !=
            for item in v.value:
                res.append(operators.not_equal(item).encode(k))
        else:
            res.append(v.encode(k))
    return ','.join(res)



from setuptools import setup
from pathlib import Path

setup(
    name='lightkube',
    version="0.10.2",
    description='Lightweight kubernetes client library',
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    author='Giuseppe Tribulato',
    author_email='gtsystem@gmail.com',
    license='MIT',
    url='https://github.com/gtsystem/lightkube',
    packages=['lightkube', 'lightkube.config', 'lightkube.core'],
    package_data={'lightkube': ['py.typed']},
    install_requires=[
        'lightkube-models >= 1.15.12.0',
        'httpx >= 0.20.0',
        'PyYAML',
        'backports-datetime-fromisoformat;python_version<"3.7"',
        'dataclasses;python_version<"3.7"'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ]
)



import sys
import typing
from typing import get_type_hints
from datetime import datetime
import dataclasses as dc

try:
    fromisoformat = datetime.fromisoformat
except AttributeError:  # python 3.6
    from backports._datetime_fromisoformat import datetime_fromisoformat as fromisoformat


from .typing_extra import get_args, get_origin

if sys.version_info[:2] > (3, 6):
    list_type = list
else:
    list_type = typing.List


def to_datetime(string):
    return fromisoformat(string.replace("Z", "+00:00"))


def from_datetime(dt):
    return dt.isoformat().replace("+00:00", "Z")


class ConverterFunc(typing.NamedTuple):
    from_json_type: typing.Callable
    to_json_type: typing.Callable


TYPE_CONVERTERS = {
    datetime: ConverterFunc(from_json_type=to_datetime, to_json_type=from_datetime)
}

EMPTY_DICT = {}


class Converter(typing.NamedTuple):
    is_list: bool
    supp_kw: bool
    func: typing.Callable

    def __call__(self, value, kw):
        if not self.supp_kw:
            kw = EMPTY_DICT
        if self.is_list:
            f = self.func
            return [f(_, **kw) for _ in value]
        return self.func(value, **kw)


def nohop(x, kw):
    return x


def is_dataclass_json(cls):
    return dc.is_dataclass(cls) and issubclass(cls, DataclassDictMixIn)


def extract_types(cls, is_to=True):
    func_name = "to_json_type" if is_to else "from_json_type"
    method_name = "to_dict" if is_to else "from_dict"
    types = get_type_hints(cls)
    for field in dc.fields(cls):
        k = field.name
        t = types[k]

        if get_origin(t) is list_type:
            is_list = True
            t = get_args(t)[0]
        else:
            is_list = False

        if is_dataclass_json(t):
            yield k, Converter(is_list=is_list, supp_kw=True, func=getattr(t, method_name)), field.default
        elif t in TYPE_CONVERTERS:
            yield k, Converter(is_list=is_list, supp_kw=False, func=getattr(TYPE_CONVERTERS[t], func_name)), field.default
        else:
            if is_to:
                yield k, nohop, field.default


class LazyAttribute:
    def __init__(self, key, convert):
        self.key = key
        self.convert = convert

    def __get__(self, instance, owner):
        value = instance._lazy_values[self.key]
        if value is not None:
            value = self.convert(value, instance._lazy_kwargs)
        setattr(instance, self.key, value)
        del instance._lazy_values[self.key]
        return value


class DataclassDictMixIn:
    _late_init_from: typing.List = None
    _late_init_to: typing.List = None
    _json_to_prop: typing.Dict = None
    _prop_to_json: typing.Dict = None
    _valid_params: typing.Set = None

    @classmethod
    def _setup(cls):
        if cls._late_init_from is None:
            cls._late_init_from = list(t[:2] for t in extract_types(cls, is_to=False))
            for k, convert in cls._late_init_from:
                setattr(cls, k, LazyAttribute(k, convert))
            cls._prop_to_json = {field.name: field.metadata['json'] for field in dc.fields(cls) if 'json' in field.metadata}
            cls._json_to_prop = {v: k for k, v in cls._prop_to_json.items()}
            cls._late_init_to = list(extract_types(cls, is_to=True))
            cls._valid_params = {f.name for f in dc.fields(cls)}

    @classmethod
    def from_dict(cls, d, lazy=True):
        cls._setup()
        kwargs = dict(lazy=lazy)
        params = cls._valid_params
        valid_d = {}
        transform = cls._json_to_prop.get
        for k, v in d.items():
            k = transform(k, k)
            if k in params:
                valid_d[k] = v
        obj = cls(**valid_d)
        if lazy:
            obj._lazy_values = {}
            obj._lazy_kwargs = kwargs
            for k, _ in cls._late_init_from:
                obj._lazy_values[k] = getattr(obj, k)
                delattr(obj, k)
        else:
            d = obj.__dict__
            for k, convert in cls._late_init_from:
                if d[k] is not None:
                    d[k] = convert(d[k], kwargs)
        return obj

    def to_dict(self, dict_factory=dict):
        self._setup()
        kwargs = dict(dict_factory=dict_factory)
        result = []
        lazy_attr = getattr(self, "_lazy_values", None)
        key_transform = self._prop_to_json.get
        for k, conv_f, default in self._late_init_to:
            if lazy_attr is not None and k in lazy_attr:
                value = lazy_attr[k]
            else:
                value = getattr(self, k)
                if value == default:
                    continue
                value = conv_f(value, kwargs)
            if value is not None:
                result.append((key_transform(k, k), value))
        return dict_factory(result)
